#include <windows.h>
#include <tchar.h>
#include <assert.h>

#include "board.h"
#include "region.h"
#include "aa.h"
#include "util.h"

/*
 * copy functions
 */

static const TCHAR *g_mb_numchar[] = {
	_T("@ "), _T("P "), _T("Q "), _T("R "), _T("S "),
	_T("T "), _T("U "), _T("V "), _T("W "), _T("X "),
};
static const TCHAR *g_mb_fixed_numchar[] = {
	_T("@"), _T("P"), _T("Q"), _T("R"), _T("S"),
	_T("T"), _T("U"), _T("V"), _T("W"), _T("X"),
};
static const TCHAR g_mb_topaxis[] = 
	_T("@ A@@B@@C@@D@@E@@F@@G@@H@@ I\r\n");
static const TCHAR g_mb_fixed_topaxis[] = 
	_T("@A@ B@ C@ D@ E@ F@ G@ H@ I\r\n");
static const LPCTSTR g_numchar[] = {
		_T("1"), _T("2"), _T("3"), _T("4"), _T("5"),
		_T("6"), _T("7"), _T("8"), _T("9"),
};

/*
      1
      
   24
      
      8
 */
static const TCHAR *g_mb_joint_upper_left[] = {
	_T(""), _T("@"), _T("@"), _T(""), _T("@"), _T(""), _T(""), _T(""),
	_T("@"), _T(""), _T(""), _T(""), _T(""), _T(""), _T(""), _T(""),
};
static const TCHAR *g_mb_joint_upper[] = {
	_T(""), _T("@"), _T("@"), _T(""), _T("@"), _T(""), _T(""), _T(""),
	_T("@"), _T(""), _T(""), _T(""), _T(""), _T(""), _T(""), _T(""),
};
static const TCHAR *g_mb_joint_upper_right[] = {
	_T(""), _T("@"), _T("@"), _T(""), _T("@"), _T(""), _T(""), _T(""),
	_T("@"), _T(""), _T(""), _T(""), _T(""), _T(""), _T(""), _T(""),
};
static const TCHAR *g_mb_joint_left[] = {
	_T(""), _T("@"), _T("@"), _T(""), _T("@"), _T(""), _T(""), _T(""),
	_T("@"), _T(""), _T(""), _T(""), _T(""), _T(""), _T(""), _T(""),
};
static const TCHAR *g_mb_joint_cross[] = {
	_T(""), _T("@"), _T("@"), _T(""), _T("@"), _T(""), _T(""), _T(""),
	_T("@"), _T(""), _T(""), _T(""), _T(""), _T(""), _T(""), _T(""),
};
static const TCHAR *g_mb_joint_right[] = {
	_T(""), _T("@"), _T("@"), _T(""), _T("@"), _T(""), _T(""), _T(""),
	_T("@"), _T(""), _T(""), _T(""), _T(""), _T(""), _T(""), _T(""),
};
static const TCHAR *g_mb_joint_bottom_left[] = {
	_T(""), _T("@"), _T("@"), _T(""), _T("@"), _T(""), _T(""), _T(""),
	_T("@"), _T(""), _T(""), _T(""), _T(""), _T(""), _T(""), _T(""),
};
static const TCHAR *g_mb_joint_bottom[] = {
	_T(""), _T("@"), _T("@"), _T(""), _T("@"), _T(""), _T(""), _T(""),
	_T("@"), _T(""), _T(""), _T(""), _T(""), _T(""), _T(""), _T(""),
};
static const TCHAR *g_mb_joint_bottom_right[] = {
	_T(""), _T("@"), _T("@"), _T(""), _T("@"), _T(""), _T(""), _T(""),
	_T("@"), _T(""), _T(""), _T(""), _T(""), _T(""), _T(""), _T(""),
};

static const WCHAR g_mb_joint_upper_left_s[] = L"";
static const WCHAR g_mb_joint_left_s[] = L"";


static void append_cross(Board *b, LPTSTR str, int x, int y, const TCHAR **l)
{
	int xx, yy;
	unsigned int v = 0;
	
	xx = (x + 1) % 9;
	yy = (y + 1) % 9;
	v |= (b->i[y][x] != b->i[y][xx]) ? 0x1 : 0;
	v |= (b->i[y][x] != b->i[yy][x]) ? 0x2 : 0;
	v |= (b->i[y][xx] != b->i[yy][xx]) ? 0x4 : 0;
	v |= (b->i[yy][x] != b->i[yy][xx]) ? 0x8 : 0;
	lstrcat(str, l[v]);
}

static void board2aamb_jigsaw(Board *b, LPTSTR str, BOOL fixed)
{
	LPCTSTR *n;
	int x, y;
	const TCHAR **l, **c, **r;
	
	if (fixed) {
		n = g_mb_fixed_numchar;
		lstrcpy(str, _T(g_mb_fixed_topaxis));
	} else {
		n = g_mb_numchar;
		lstrcpy(str, _T(g_mb_topaxis));
	}
	append_cross(b, str, 8, 8, g_mb_joint_upper_left);
	for (x = 0; x < 8; x ++) {
		lstrcat(str, (b->i[0][x] == b->i[8][x]) ? _T("") : _T(""));
		append_cross(b, str, x, 8, g_mb_joint_upper);
	}
	lstrcat(str, (b->i[0][8] == b->i[8][8]) ? _T("") : _T(""));
	append_cross(b, str, 8, 8, g_mb_joint_upper_right);
	lstrcat(str, _T("\r\n"));
	for (y = 0; y < 9; y ++) {
		lstrcat(str, (b->i[y][0] == b->i[y][8]) ? _T("") : _T(""));
		for (x = 0; x < 9; x ++) {
			lstrcat(str, n[bit2int(b->n[y][x])]);
			lstrcat(str, (b->i[y][x] == b->i[y][(x + 1) % 9]) ? _T("") : _T(""));
		}
		lstrcat(str, g_numchar[y]);
		lstrcat(str, _T("\r\n"));
		if (y < 8) {
			l = g_mb_joint_left; c = g_mb_joint_cross; r = g_mb_joint_right;
		} else {
			l = g_mb_joint_bottom_left; c = g_mb_joint_bottom; r = g_mb_joint_bottom_right;
		}
		append_cross(b, str, 8, y, l);
		for (x = 0; x < 8; x ++) {
			lstrcat(str, (b->i[y][x] != b->i[(y + 1) % 9][x]) ? _T("") : _T(""));
			append_cross(b, str, x, y, c);
		}
		lstrcat(str, (b->i[y][8] != b->i[(y + 1) % 9][8]) ? _T("") : _T(""));
		append_cross(b, str, 8, y, r);
		lstrcat(str, _T("\r\n"));
	}
}

static void board2aamb_normal(Board *b, LPTSTR str, BOOL fixed)
{
	LPCTSTR *n;
	int x, y;
	
	if (fixed) {
		n = g_mb_fixed_numchar;
		lstrcpy(str, _T(g_mb_fixed_topaxis));
	} else {
		n = g_mb_numchar;
		lstrcpy(str, _T(g_mb_topaxis));
	}
	lstrcat(str, _T("\r\n"));
	for (y = 0; y < 9; y ++) {
		if (y % 3 == 0) {
			if (y > 0) {
				lstrcat(str, _T("\r\n"));
			}
		} else {
			lstrcat(str, _T("\r\n"));
		}
		for (x = 0; x < 9; x ++) {
			if (x % 3 == 0) {
				lstrcat(str, _T(""));
			} else {
				lstrcat(str, _T(""));
			}
			lstrcat(str, n[bit2int(b->n[y][x])]);
		}
		lstrcat(str, "");
		lstrcat(str, g_numchar[y]);
		lstrcat(str, _T("\r\n"));
	}
	lstrcat(str, _T("\r\n"));
}

void board2aamb(Board *b, LPTSTR str)
{
	if (is_jigsaw(b)) {
		board2aamb_jigsaw(b, str, FALSE);
	} else {
		board2aamb_normal(b, str, FALSE);
	}
}

void board2aamb_fixed(Board *b, LPTSTR str)
{
	if (is_jigsaw(b)) {
		board2aamb_jigsaw(b, str, TRUE);
	} else {
		board2aamb_normal(b, str, TRUE);
	}
}

void board2aa_normal(Board *b, LPTSTR str)
{
	static const TCHAR n[][3] = {
		_T(". "), _T("1 "), _T("2 "), _T("3 "), _T("4 "),
		_T("5 "), _T("6 "), _T("7 "), _T("8 "), _T("9 "),
	};
	int x, y;
	
	str[0] = _T('\0');
	for (y = 0; y < 9; y ++) {
		if (y % 3 == 0) {
			lstrcat(str, _T("+-------+-------+-------+\r\n"));
		}
		for (x = 0; x < 9; x ++) {
			if (x % 3 == 0) {
				lstrcat(str, _T("| "));
			}
			lstrcat(str, n[bit2int(b->n[y][x])]);
		}
		lstrcat(str, _T("|\r\n"));
	}
	lstrcat(str, _T("+-------+-------+-------+\r\n"));
}


/*
 * paste functions
 */

static int split_region_fill(Board *b, int right[9][9], int bottom[9][9],
                             int x, int y, int n)
{
	int c = 0, l;
	
	l = (x + 8) % 9;
	if (right[y][l] == 0 && b->i[y][l] == 0) {
		b->i[y][l] = n;
		c ++;
		c += split_region_fill(b, right, bottom, l, y, n);
	}
	l = (x + 1) % 9;
	if (right[y][x] == 0 && b->i[y][l] == 0) {
		b->i[y][l] = n;
		c ++;
		c += split_region_fill(b, right, bottom, l, y, n);
	}
	l = (y + 8) % 9;
	if (bottom[l][x] == 0 && b->i[l][x] == 0) {
		b->i[l][x] = n;
		c ++;
		c += split_region_fill(b, right, bottom, x, l, n);
	}
	l = (y + 1) % 9;
	if (bottom[y][x] == 0 && b->i[l][x] == 0) {
		b->i[l][x] = n;
		c ++;
		c += split_region_fill(b, right, bottom, x, l, n);
	}
	return c;
}

static int split_region(Board *b, int right[9][9], int bottom[9][9])
{
	int i, c, x, y;
	
	i = 1;
	for (y = 0; y < 9; y ++) {
		for (x = 0; x < 9; x ++) {
			if (b->i[y][x] == 0) {
				b->i[y][x] = i;
				c = split_region_fill(b, right, bottom, x, y, i) + 1;
				trace("%d %d", i, c);
				i ++;
				if (c != 9) return 1;
			}
		}
	}
	return (i != 10) ? 1 : 0;
}

static BOOL set_board(Board *b, int n[9][9], int right[9][9], int bottom[9][9])
{
	int i, j;
	
	zero_board(b);
	if (split_region(b, right, bottom)) return FALSE;
	for (i = 0; i < 9; i ++) {
		for (j = 0; j < 9; j ++) {
			b->i[i][j] --;
		}
	}
	normalize_region(b);
	make_region_index(b);
	for (i = 0; i < 9; i ++) {
		trace("%d %d %d %d %d %d %d %d %d ", 
		      b->i[i][0], b->i[i][1], b->i[i][2], b->i[i][3], b->i[i][4], 
		      b->i[i][5], b->i[i][6], b->i[i][7], b->i[i][8]);
		for (j = 0; j < 9; j ++) {
			if (n[i][j]) {
				if (put_number(b, j, i, n[i][j])) {
					return FALSE;
				}
			}
		}
	}
	assert(check_region(b) == 0);
	return TRUE;
}

static LPCWSTR skip_for(LPCWSTR p, LPCWSTR tbl)
{
	LPCWSTR t;
	
	for (; *p != L'\0'; p ++) {
		for (t = tbl; *t != L'\0'; t ++) {
			if (*t == *p) return p + 1;
		}
	}
	return NULL;
}

static BOOL line_aa2board(Board *b, LPCWSTR wstr)
{
	int right[9][9] = {{0}, {0}}, bottom[9][9] = {{0}, {0}}, n[9][9] = {{0}, {0}};
	int i, j;
	static const WCHAR nl[] = L"\r\n";
	LPCWSTR p;
	
	p = skip_for(wstr, g_mb_joint_upper_left_s);
	if (! p) return FALSE;
	for (i = 0; i < 9; i ++) {
		p = skip_for(p, nl);
		if (! p) return FALSE;
		p = skip_for(p, L"b");
		if (! p) return FALSE;
		for (j = 0; j < 9; j ++) {
			for (; *p != L'' && *p != L'' && *p != L'b'; p ++) {
				if (*p >= L'O' && *p <= L'X') {
					n[i][j] = *p - L'O';
				} else if (*p >= L'0' && *p <= L'9') {
					n[i][j] = *p - L'0';
				} else if (*p == L'\0') {
					return FALSE;
				}
			}
			right[i][j] = (*p == L'') ? 1 : 0;
			p ++;
		}
		p = skip_for(p, nl);
		if (! p) return FALSE;
		p = skip_for(p, g_mb_joint_left_s);
		if (! p) return FALSE;
		for (j = 0; j < 9; j ++) {
			if (*p == L'' || *p == L'\') {
				bottom[i][j] = 0;
			} else if (*p == L'') {
				bottom[i][j] = 1;
			} else {
				return FALSE;
			}
			if (* ++p == L'\0') return FALSE;
			p ++;
		}
	}
	return set_board(b, n, right, bottom);
}

static BOOL aa_sc2board(Board *b, LPCWSTR str, LPCWSTR spaces, BOOL paren)
{
	int i;
	LPCWSTR p, q;
	
	initialize_board(b);
	for (i = 0, p = str; i < 9 * 9 && *p; p ++) {
		if (paren && (*p == L'(' || *p == L'i')) {
			for (; *p != L')' && *p != L'j'; p ++) {
				if (*p == L'\0') return FALSE;
			}
			continue;
		}
		for (q = spaces; *q && *p != *q; q ++)
			;
		if (*q) {
			i ++;
		} else if (*p >= L'P' && *p <= L'X') {
			if (put_number(b, i % 9, i / 9, *p - L'O')) {
				return FALSE;
			}
			i ++;
		}
	}
	return i == 81;
}

BOOL aa2board(Board *b, LPCTSTR str)
{
	LPCWSTR wstr = NULL;
	BOOL result = FALSE;
#ifdef _UNICODE
	wstr = str;
#else
	{
		int length, size;
		
		length = MultiByteToWideChar(CP_ACP, 0, str, -1, NULL, 0);
		if (length == 0) goto RESULT;
		size = length * sizeof(WCHAR);
		wstr = malloc(size);
		if (! wstr) goto RESULT;
		length = MultiByteToWideChar(CP_ACP, 0, str, -1,
		                             (LPWSTR) wstr, length);
		if (length == 0) goto RESULT;
	}
#endif
	{
		WCHAR p[2] = {0};
		LPCWSTR pp;
		
		for (pp = L"ODQ|H{E"; *pp != L'\0'; pp ++) {
			p[0] = *pp;
			if (aa_sc2board(b, wstr, p, FALSE)) {
				result = TRUE;
				goto RESULT;
			}
		}
	}
	if (line_aa2board(b, wstr)) {
		result = TRUE;
		goto RESULT;
	}
	if (aa_sc2board(b, wstr, L"@", FALSE)) {
		result = TRUE;
		goto RESULT;
	}
RESULT:
#ifndef _UNICODE
	free((LPWSTR) wstr);
#endif
	return result;
}

