#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "board.h"
#include "gen.h"
#include "print.h"
#include "random.h"
#include "symmetry.h"

#if defined(_WIN32) && ! defined(__CONSOLE__)
#include <windows.h>
#include "util.h"
int is_canceled(void);
#endif

void shuffle(unsigned int *a, int n)
{
	unsigned int b, i, j;
	
	for (i = n; i > 1; i --, a ++) {
//		j = rand() / ((RAND_MAX + i) / i);
		j = range_rand(i);
		assert(j < i);
		b = a[0];  a[0] = a[j];  a[j] = b;
	}
}

#if 0

int generate(Board *b, int x, int y)
{
	unsigned int n, m, p[9];
	int i, j;
	
	if (++ x == 9 && (x = 0, ++ y == 9)) {
		return 1;
	}
	if (b->n[y][x]) {
		return generate(b, x, y);
	}
	n = ~(b->x[x] | b->y[y] | b->b[b->i[y][x]]) & 0x1ff;
	i = 0;
	while (n != 0) {
		m = n & -n;
		assert(m != 0);
		p[i ++] = m;
		n &= ~m;
	}
	assert(i <= 9);
	shuffle(p, i);
	for (j = 0; j < i; j ++) {
		put(b, x, y, p[j]);
		if (generate(b, x, y)) {
			return 1;
		}
		rm(b, x, y);
	}
	return 0;
}

#else

static int generate_(Board *);

static int decisive_fill(Board *b)
{
	unsigned int n, m;
	int c, i, j, r, x, y, xx, yy;
	
	for (y = 0; y < 9; y ++) {
		for (x = 0; x < 9; x ++) {
			if (b->n[y][x]) continue;
			n = ~(b->x[x] | b->y[y] | b->b[b->i[y][x]]) & 0x1ff;
			c = 0;
			while (n != 0) {
				m = n & -n; n &= ~m; c ++;
			}
			switch (c) {
			case 0:
				return 1;
			case 1:
				put(b, x, y, m);
				r = generate_(b);
				if (r < 0) return r;
				rm(b, x, y);
				return 1;
			}
		}
	}
	
	for (y = 0; y < 9; y ++) {
		n = ~(b->y[y]) & 0x1ff;
		while (n != 0) {
			m = n & -n;
			assert(m != 0);
			c = 0;
			for (x = 0; x < 9; x ++) {
				if (! b->n[y][x] &&
				    (~(b->x[x] | b->b[b->i[y][x]]) & m) != 0) {
					c ++; xx = x;
				}
			}
			switch (c) {
			case 0:
				return 1;
			case 1:
				put(b, xx, y, m);
				r = generate_(b);
				if (r < 0) return r;
				rm(b, xx, y);
				return 1;
			}
			n &= ~m;
		}
	}
	
	for (x = 0; x < 9; x ++) {
		n = ~(b->x[x]) & 0x1ff;
		while (n != 0) {
			m = n & -n;
			assert(m != 0);
			c = 0;
			for (y = 0; y < 9; y ++) {
				if (! b->n[y][x] &&
				    (~(b->y[y] | b->b[b->i[y][x]]) & m) != 0) {
					c ++; yy = y;
				}
			}
			switch (c) {
			case 0:
				return 1;
			case 1:
				put(b, x, yy, m);
				r = generate_(b);
				if (r < 0) return r;
				rm(b, x, yy);
				return 1;
			}
			n &= ~m;
		}
	}
	
	for (i = 0; i < 9; i ++) {
		n = ~(b->b[i]) & 0x1ff;
		while (n != 0) {
			m = n & -n;
			assert(m != 0);
			c = 0;
			for (j = 0; j < 9; j ++) {
				x = b->j[i][j] >> 4;
				y = b->j[i][j] & 0xf;
				if (! b->n[y][x] &&
				    (~(b->y[y] | b->x[x]) & m) != 0) {
					c ++; xx = x; yy = y;
				}
			}
			switch (c) {
			case 0:
				return 1;
			case 1:
				put(b, xx, yy, m);
				r = generate_(b);
				if (r < 0) return r;
				rm(b, xx, yy);
				return 1;
			}
			n &= ~m;
		}
	}
	return 0;
}

static int generate_(Board *b)
{
	unsigned int n, m, p[9];
	int i, j, x, y, r;
	
#if defined(_WIN32) && ! defined(__CONSOLE__)
	if (is_canceled()) return -2;
#endif
	r = decisive_fill(b);
	if (r < 0) return r;
	if (r) return 0;
	if (! search_minimum_choice(b, &x, &y)) return -1;
	n = ~(b->x[x] | b->y[y] | b->b[b->i[y][x]]) & 0x1ff;
	i = 0;
	while (n != 0) {
		m = n & -n;
		assert(m != 0);
		p[i ++] = m;
		n &= ~m;
	}
	assert(i <= 9);
	shuffle(p, i);
	for (j = 0; j < i; j ++) {
		put(b, x, y, p[j]);
		r = generate_(b);
		if (r) return r;
		rm(b, x, y);
	}
	return 0;
}

int generate(Board *b, int x, int y)
{
	return generate_(b);
}
#endif

int random_scan_rm(Board *b)
{
	unsigned int pos[9 * 9], v;
	int i, x, y;
	
	for (i = 0; i < 9 * 9; i ++) {
		pos[i] = ((i % 9) << 8) | (i / 9);
	}
	shuffle(pos, 9 * 9);
	for (i = 0; i < 9 * 9; i ++) {
#if defined(_WIN32) && ! defined(__CONSOLE__)
		if (is_canceled()) return 1;
#endif
		x = pos[i] >> 8;
		y = pos[i] & 0xf;
		v = b->n[y][x];
		assert(x >= 0 && x < 9 && y >= 0 && y < 9);
		assert(v != 0);
		rm(b, x, y);
		b->sol = 0;
		if (search(b, -1, 0) < -1) return 1;
		if (b->sol != 1) {
			put(b, x, y, v);
		}
	}
	return 0;
}


int random_scan_rm_s(Board *b, int p)
{
	unsigned int pos[9 * 9], v = 0, v2 = 0;
	int i, n, x = 0, y = 0, x2 = 0, y2 = 0;
	
	n = 0;
	switch (p) {
	case SYM_H:
		for (y = 0; y < 9; y ++) {
			for (x = 0; x < 5; x ++) {
				pos[n ++] = (x << 8) | y;
			}
		}
		break;
	case SYM_V:
		for (y = 0; y < 5; y ++) {
			for (x = 0; x < 9; x ++) {
				pos[n ++] = (x << 8) | y;
			}
		}
		break;
	case SYM_P:
		for (y = 0; y < 4; y ++) {
			for (x = 0; x < 9; x ++) {
				pos[n ++] = (x << 8) | y;
			}
		}
		for (x = 0; x < 5; x ++) {
			pos[n ++] = (x << 8) | 4;
		}
		break;
	default:
		for (y = 0; y < 9; y ++) {
			for (x = 0; x < 9; x ++) {
				pos[n ++] = (x << 8) | y;
			}
		}
		break;
	}
	shuffle(pos, n);
	for (i = 0; i < n; i ++) {
#if defined(_WIN32) && ! defined(__CONSOLE__)
		if (is_canceled()) return 1;
#endif
		x = pos[i] >> 8;
		y = pos[i] & 0xf;
		v = b->n[y][x];
		rm(b, x, y);
		if (p != SYM_N) {
			symmetry(x, y, &x2, &y2, p);
			if (x != x2 || y != y2) {
				v2 = b->n[y2][x2];
				rm(b, x2, y2);
			}
		}
		b->sol = 0;
		if (search(b, -1, 0) < -1) return 1;
		if (b->sol != 1) {
			put(b, x, y, v);
			if (p != SYM_N && (x != x2 || y != y2)) {
				put(b, x2, y2, v2);
			}
		}
	}
	return 0;
}
