#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include <assert.h>

#include "board.h"
#include "random.h"
#include "symmetry.h"

static void print(int b[9][9])
{
	int x, y, n;
	
	for (y = 0; y < 9; y ++) {
		for (x = 0; x < 9; x ++) {
			n = b[y][x];
			printf("%d", n);
		}
		putchar('\n');
	}
}

static int check_split_fill(int b[9][9], int f[9][9],
                            int x, int y, int n)
{
	int c = 0, l;
	
	l = (x + 8) % 9;
	if (b[y][l] == n && f[y][l] == 0) {
		f[y][l] = 1;
		c ++;
		c += check_split_fill(b, f, l, y, n);
	}
	l = (x + 1) % 9;
	if (b[y][l] == n && f[y][l] == 0) {
		f[y][l] = 1;
		c ++;
		c += check_split_fill(b, f, l, y, n);
	}
	
	l = (y + 8) % 9;
	if (b[l][x] == n && f[l][x] == 0) {
		f[l][x] = 1;
		c ++;
		c += check_split_fill(b, f, x, l, n);
	}
	l = (y + 1) % 9;
	if (b[l][x] == n && f[l][x] == 0) {
		f[l][x] = 1;
		c ++;
		c += check_split_fill(b, f, x, l, n);
	}
	return c;
}

static int check_split(int b[9][9])
{
	int i, c, x, y;
	int f[9][9];
	
	for (i = 0; i < 9; i ++) {
		for (y = 0; y < 9; y ++) {
			for (x = 0; x < 9; x ++) {
				if (b[y][x] == i) {
					memset(f, 0, sizeof(f));
					f[y][x] = 1;
					c = check_split_fill(b, f, x, y, i) + 1;
					if (c != 9) return 1;
					goto label;
				}
			}
		}
		return 1;
		label: ;
	}
	return 0;
}

static int check_remove(int b[9][9], int x, int y)
{
	unsigned int n[3][3];
	
	/*
	 * 012
	 * 1x
	 * 2
	 */
	n[0][0] = (y > 0 && x > 0) ? b[y - 1][x - 1] : -1;
	n[0][1] = (y > 0) ? b[y - 1][x] : -1;
	n[0][2] = (y > 0 && x < 8) ? b[y - 1][x + 1] : -1;
	n[1][0] = (x > 0) ? b[y][x - 1] : -1;
	n[1][1] = b[y][x];
	n[1][2] = (x < 8) ? b[y][x + 1] : -1;
	n[2][0] = (y < 8 && x > 0) ? b[y + 1][x - 1] : -1;
	n[2][1] = (y < 8) ? b[y + 1][x] : -1;
	n[2][2] = (y < 8 && x < 8) ? b[y + 1][x + 1] : -1;
	
	/*
	 *  o
	 * ox
	 */
	if (n[1][1] == n[0][1] && n[1][1] == n[1][0] && n[1][1] != n[0][0]) {
		return 0;
	}
	/*
	 * ox
	 *  o
	 */
	if (n[1][1] == n[0][1] && n[1][1] == n[1][2] && n[1][1] != n[0][2]) {
		return 0;
	}
	/*
	 * o
	 * xo
	 */
	if (n[1][1] == n[2][1] && n[1][1] == n[1][0] && n[1][1] != n[2][0]) {
		return 0;
	}
	/*
	 * xo
	 * o
	 */
	if (n[1][1] == n[2][1] && n[1][1] == n[1][2] && n[1][1] != n[2][2]) {
		return 0;
	}
	/*
	 * oxo
	 */
	if (n[1][1] == n[1][0] && n[1][1] == n[1][2] &&
	    (n[1][1] != n[0][0] || n[1][1] != n[0][1] || n[1][1] != n[0][2]) &&
	    (n[1][1] != n[2][0] || n[1][1] != n[2][1] || n[1][1] != n[2][2])) {
		return 0;
	}
	/*
	 * o
	 * x
	 * o
	 */
	if (n[1][1] == n[0][1] && n[1][1] == n[2][1] &&
	    (n[1][1] != n[0][0] || n[1][1] != n[1][0] || n[1][1] != n[2][0]) &&
	    (n[1][1] != n[0][2] || n[1][1] != n[1][2] || n[1][1] != n[2][2])) {
		return 0;
	}

	/*
	 * ooo
	 * o o doesn't appear. because size of void is less than 9.
	 * ooo
	 */
	
	return 1;
}

struct fill {
	int x, y, n, m;
}; 

static void shuffle(struct fill *a, int n)
{
	struct fill b;
	unsigned int i, j;
	
	for (i = n; i > 1; i --, a ++) {
//		j = rand() / ((RAND_MAX + i) / i);
		j = range_rand(i);
		assert(j < i);
		b = a[0];  a[0] = a[j];  a[j] = b;
	}
}

static int generate(int b[9][9], struct fill *l, int r, int s)
{
	static const struct {
		int x, y;
	} d[] = {{-1, 0}, {1, 0}, {0, -1}, {0, 1}};
	struct fill p[20];
	int x, y, sx, sy, xx, yy, sxx, syy;
	int i, n, o;
	
	if (l->n == 0 && r != 0) {
		assert(check_split(b) == 0);
		return 1;
	}
	n = 0;
	for (y = 0; y < 9; y ++) {
		for (x = 0; x < 9; x ++) {
			if (s) {
				symmetry(x, y, &sx, &sy, s);
				if (x == sx && y == sy) continue;
			}
			if (b[y][x] != l->n || ! check_remove(b, x, y)) continue;
			for (i = 0; i < 4; i ++) {
				xx = x + d[i].x;
				yy = y + d[i].y;
				if (xx < 0 || xx >= 9 || yy < 0 || yy >= 9) continue;
				if (s) {
					symmetry(xx, yy, &sxx, &syy, s);
					if (sxx < 0 || sxx >= 9 || syy < 0 || syy >= 9) continue;
				}
				if (b[y][x] != b[yy][xx] &&
				    (b[yy][xx] != r || l->x != x || l->y != y)) {
					p[n].x = x;
					p[n].y = y;
					p[n].n = b[yy][xx];
					if (s) {
						p[n].m = b[syy][sxx];
					}
					n ++;
					assert(n < 20);
				}
			}
		}
	}
	if (n > 0) {
		shuffle(p, n);
		for (i = 0; i < n; i ++) {
			b[p[i].y][p[i].x] = p[i].n;
			if (s) {
				symmetry(p[i].x, p[i].y, &sx, &sy, s);
				o = b[sy][sx];
				b[sy][sx] = p[i].m;
			}
			if (generate(b, &p[i], l->n, s)) {
				return 1;
			}
			b[p[i].y][p[i].x] = l->n;
			if (s) {
				b[sy][sx] = o;
			}
		}
	}
	return 0;
}

void normalize_region(Board *b)
{
	unsigned int u, v;
	int n[9], c, x, y, i;
	
	c = 0;
	u = 0;
	for (y = 0; y < 9; y ++) {
		for (x = 0; x < 9; x ++) {
			i = b->i[y][x];
			v = 0x1 << i;
			if (u & v) continue;
			u |= v;
			n[i] = c ++;
		}
	}
	for (y = 0; y < 9; y ++) {
		for (x = 0; x < 9; x ++) {
			b->i[y][x] = n[b->i[y][x]];
		}
	}
}

void make_region_index(Board *b)
{
	int n[9], i, x, y;
	
	for (i = 0; i < 9; i ++) n[i] = 0;
	for (y = 0; y < 9; y ++) {
		for (x = 0; x < 9; x ++) {
			i = b->i[y][x];
			b->j[i][n[i] ++] = ((0xf & x) << 4) | (0xf & y);
		}
	}
}

static const int initial_region[9][9] = {
	{0,0,0,1,1,1,2,2,2,},
	{0,0,0,1,1,1,2,2,2,},
	{0,0,0,1,1,1,2,2,2,},
	{3,3,3,4,4,4,5,5,5,},
	{3,3,3,4,4,4,5,5,5,},
	{3,3,3,4,4,4,5,5,5,},
	{6,6,6,7,7,7,8,8,8,},
	{6,6,6,7,7,7,8,8,8,},
	{6,6,6,7,7,7,8,8,8,},
};

int generate_region(Board *b, int g, int s)
{
	struct fill p = {-1, -1, 0};
	int i, r = 1;
	
	memcpy(b->i, initial_region, sizeof(initial_region));
	if (g) {
		for (i = 0; i < 50; i ++) {
			r = generate(b->i, &p, 0, s);
			if (!r) break;
		}
		normalize_region(b);
	}
	make_region_index(b);
	return r;
}

int is_jigsaw(Board *b)
{
	return memcmp(b->i, initial_region, sizeof(initial_region));
}

int check_region(Board *b)
{
	return check_split(b->i);
}
