#include <assert.h>

#ifdef _WIN32
#include <windows.h>
#include <tchar.h>
/* for trace */
#include "util.h"
#else
#define _T(x) x
typedef char TCHAR;
typedef char *LPTSTR;
typedef const char *LPCTSTR;
#endif

#include "board.h"
#include "strw.h"
#include "region.h"
#include "aa.h"

void board2code(Board *ini, Board *cur, LPTSTR str)
{
	static const TCHAR hexl[] = _T("abcdefghijk");
	static const TCHAR hexu[] = _T("ABCDEFGHIJK");
	int i, x, y;
	
	i = 0;
	for (y = 0; y < 9; y ++) {
		for (x = 0; x < 9; x ++) {
			if (ini->n[y][x]) {
				str[i ++] = hexl[bit2int(ini->n[y][x])];
			} else {
				str[i ++] = hexu[bit2int(cur->n[y][x])];
			}
		}
	}
	str[i] = _T('\0');
}

int code2board(Board *ini, Board *cur, LPTSTR str)
{
	int i, x, y, n;
	TCHAR c;
	
	i = 0;
	for (y = 0; y < 9; y ++) {
		for (x = 0; x < 9; x ++) {
			c = str[i ++];
			if (c >= _T('a') && c <= _T('j')) {
				n = c - _T('a');
				if (put_number(ini, x, y, n) ||
				    put_number(cur, x, y, n)) {
					return 0;
				}
			} else if (c >= _T('A') && c <= _T('J')) {
				if (put_number(cur, x, y, c - _T('A'))) {
					return 0;
				}
			} else {
				return 0;
			}
		}
	}
	return 1;
}

void region2code(Board *b, LPTSTR str)
{
	static const TCHAR region[] = _T("012345678");
	int i, x, y;
	
	i = 0;
	for (y = 0; y < 9; y ++) {
		for (x = 0; x < 9; x ++) {
			assert(b->i[y][x] >= 0 && b->i[y][x] < 9);
			str[i ++] = region[b->i[y][x]];
		}
	}
	str[i] = _T('\0');
}

int code2region(Board *b, LPTSTR str)
{
	int i, x, y;
	TCHAR c;
	
	i = 0;
	zero_board(b);
	for (y = 0; y < 9; y ++) {
		for (x = 0; x < 9; x ++) {
			c = str[i ++];
			if (c >= _T('0') && c < _T('9')) {
				b->i[y][x] = c - _T('0');
			} else {
				return 0;
			}
		}
	}
	make_region_index(b);
	return 1;
}

#ifdef _WIN32

/*
 * copy functions
 */

void board2ascii(Board *b, LPTSTR str)
{
	static const TCHAR n[][2] = {
		_T("-"), _T("1"), _T("2"), _T("3"), _T("4"),
		_T("5"), _T("6"), _T("7"), _T("8"), _T("9"),
	};
	int x, y;
	
	str[0] = _T('\0');
	for (y = 0; y < 9; y ++) {
		for (x = 0; x < 9; x ++) {
			lstrcat(str, n[bit2int(b->n[y][x])]);
		}
		lstrcat(str, _T("\r\n"));
	}
}

void board2ascii_spc(Board *b, LPTSTR str)
{
	static const TCHAR n[][2] = {
		_T("-"), _T("1"), _T("2"), _T("3"), _T("4"),
		_T("5"), _T("6"), _T("7"), _T("8"), _T("9"),
	};
	int x, y;
	
	str[0] = _T('\0');
	for (y = 0; y < 9; y ++) {
		if (y > 0 && y % 3 == 0) {
			lstrcat(str, _T("\r\n"));
		}
		for (x = 0; x < 9; x ++) {
			if (x > 0 && x % 3 == 0) {
				lstrcat(str, _T(" "));
			}
			lstrcat(str, n[bit2int(b->n[y][x])]);
		}
		lstrcat(str, _T("\r\n"));
	}
}

static void board2ascii_axis_main(Board *b, LPTSTR str, BOOL fixed)
{
	static const TCHAR n[][2] = {
		_T("-"), _T("1"), _T("2"), _T("3"), _T("4"),
		_T("5"), _T("6"), _T("7"), _T("8"), _T("9"),
	};
	static const TCHAR header[] = _T("@@abc def@ghi\r\n");
	static const TCHAR header_fixed[] = _T("    abc def ghi\r\n");
	static const TCHAR left[][5] = {
		_T("(1) "), _T("(2) "), _T("(3) "), _T("(4) "), _T("(5) "), 
		_T("(6) "), _T("(7) "), _T("(8) "), _T("(9) "), 
	};
	int x, y;
	
	lstrcpy(str, fixed ? header_fixed : header);
	for (y = 0; y < 9; y ++) {
		if (y > 0 && y % 3 == 0) {
			lstrcat(str, _T("\r\n"));
		}
		lstrcat(str, left[y]);
		for (x = 0; x < 9; x ++) {
			if (x > 0 && x % 3 == 0) {
				lstrcat(str, _T(" "));
			}
			lstrcat(str, n[bit2int(b->n[y][x])]);
		}
		lstrcat(str, _T("\r\n"));
	}
}

void board2ascii_axis(Board *b, LPTSTR str)
{
	board2ascii_axis_main(b, str, FALSE);
}

void board2ascii_axis_fixed(Board *b, LPTSTR str)
{
	board2ascii_axis_main(b, str, TRUE);
}

void board2csv(Board *b, LPTSTR str)
{
	static const TCHAR n[][3] = {
		_T(""), _T("1"), _T("2"), _T("3"), _T("4"),
		_T("5"), _T("6"), _T("7"), _T("8"), _T("9"),
	};
	int x, y;
	
	str[0] = _T('\0');
	for (y = 0; y < 9; y ++) {
		for (x = 0; x < 8; x ++) {
			lstrcat(str, n[bit2int(b->n[y][x])]);
			lstrcat(str, ",");
		}
		lstrcat(str, n[bit2int(b->n[y][8])]);
		lstrcat(str, _T("\r\n"));
	}
}

static void board2early_period_arrangement(Board *b, LPTSTR str, BOOL nl)
{
	static const TCHAR n[][3] = {
		_T("-"), _T("1"), _T("2"), _T("3"), _T("4"),
		_T("5"), _T("6"), _T("7"), _T("8"), _T("9"),
	};
	int x, y;
	
	for (y = 0; y < 9; y ++) {
		lstrcat(str, _T("."));
		for (x = 0; x < 9; x ++) {
			lstrcat(str, n[bit2int(b->n[y][x])]);
		}
		if (nl) {
			lstrcat(str, _T("\r\n"));
		} else if (y < 8) {
			lstrcat(str, _T("/"));
		}
	}
}

void board2zuntan(Board *b, LPTSTR str)
{
	str[0] = _T('\0');
	board2early_period_arrangement(b, str, TRUE);
}

void board2zuntanurl(Board *b, LPTSTR str)
{
	lstrcpy(str, _T(ZUNTAN_URL));
	board2early_period_arrangement(b, str, FALSE);
}

static void board2simpleurl(Board *b, LPTSTR str)
{
	static const TCHAR n[][3] = {
		_T("0"), _T("1"), _T("2"), _T("3"), _T("4"),
		_T("5"), _T("6"), _T("7"), _T("8"), _T("9"),
	};
	int x, y;
	
	for (y = 0; y < 9; y ++) {
		for (x = 0; x < 9; x ++) {
			lstrcat(str, n[bit2int(b->n[y][x])]);
		}
	}
}

void board2brainbashers(Board *b, LPTSTR str)
{
	lstrcpy(str, _T(BRAINBASHERS_URL));
	board2simpleurl(b, str);
}

void board2playr_co_uk(Board *b, LPTSTR str)
{
	lstrcpy(str, _T(PLAYR_CO_UK_URL));
	board2simpleurl(b, str);
}

/*
 * paste functions
 */

static BOOL zuntan2board(Board *b, LPCTSTR str)
{
	int x, y;
	LPCTSTR p;
	
	initialize_board(b);
	p = str;
	for (y = 0; y < 9; y ++) {
		do {
			for (; *p != '.'; p ++) {
				if (*p == _T('\0')) return FALSE;
			}
			p ++;
		} while ((*p < _T('1') || *p > _T('9')) && *p != _T('-'));
		for (x = 0; x < 9; x ++, p ++) {
			if (*p >= _T('1') && *p <= _T('9')) {
				if (put_number(b, x, y, *p - _T('0'))) {
					return FALSE;
				}
			} else if (*p != _T('-')) return FALSE;
		}
	}
	return TRUE;
}

static BOOL sc2board(Board *b, LPCTSTR str, LPCTSTR spaces, BOOL paren)
{
	int i;
	LPCTSTR p, q;
	
	initialize_board(b);
	for (i = 0, p = str; i < 9 * 9 && *p; p ++) {
		if (paren && *p == _T('(')) {
			for (; *p != _T(')'); p ++) {
				if (*p == _T('\0')) return FALSE;
			}
			continue;
		}
		for (q = spaces; *q && *p != *q; q ++)
			;
		if (*q) {
			i ++;
		} else if (*p >= _T('1') && *p <= _T('9')) {
			if (put_number(b, i % 9, i / 9, *p - _T('0'))) {
				return FALSE;
			}
			i ++;
		}
	}
	return i == 81;
}

static BOOL sv2board(Board *b, LPCTSTR str, LPCTSTR delimiters)
{
	int i;
	LPCTSTR p, q;
	
	initialize_board(b);
	for (i = 0, p = str; i < 9 * 9 && *p; p ++) {
		for (q = delimiters; *q && *p != *q; q ++)
			;
		if (*q) {
			i ++;
		} else if (*p >= _T('1') && *p <= _T('9')) {
			if (put_number(b, i % 9, i / 9, *p - _T('0'))) {
				return FALSE;
			}
		}
	}
	return i == 81;
}

BOOL ascii2board(Board *b, LPCTSTR str)
{
	TCHAR p[3] = {0};
	LPCTSTR pp;
	
	if (! str) return FALSE;
	// AA
	if (aa2board(b, str)) {
		trace("aa");
		return TRUE;
	}
	// Early period arrangement
	if (zuntan2board(b, str)) {
		trace("zuntan");
		return TRUE;
	}
	// space charactor with ignoring in parenthesis
	for (pp = _T("0._-?=+*#"); *pp != _T('\0'); pp ++) {
		p[0] = *pp;
		if (sc2board(b, str, p, TRUE)) {
			trace("parenthesis");
			return TRUE;
		}
	}
	// space charactor
	for (pp = _T("0._-?=+*#"); *pp != _T('\0'); pp ++) {
		p[0] = *pp;
		if (sc2board(b, str, p, FALSE)) {
			trace("space");
			return TRUE;
		}
	}
	// separater
	for (pp = _T(",\t"); *pp != _T('\0'); pp ++) {
		p[0] = *pp;
		if (sv2board(b, str, p)) {
			trace("csv");
			return TRUE;
		}
	}
	// with newline
	p[1] = _T('\n');
	for (pp = _T(",\t"); *pp != _T('\0'); pp ++) {
		p[0] = *pp;
		if (sv2board(b, str, p)) {
			trace("separater");
			return TRUE;
		}
	}
	return FALSE;
}

int ascii2region(Board *b, LPTSTR str)
{
	int i, j, x, y, n[10] = {0};
	LPCTSTR p;
	
	zero_board(b);
	for (i = 0, p = str; i < 9 * 9 && *p; p ++) {
		if (*p >= _T('0') && *p <= _T('9')) {
			j = *p - _T('0');
			b->i[i / 9][i % 9] = j;
			n[j] ++;
			i ++;
		}
	}
	if (i != 81) return 0;
	for (i = 0; i < 10; i ++) {
		if (n[i] == 0) {
			for (j = i + 1; j < 10; j ++) {
				n[j] = j - 1;
			}
			break;
		} else {
			n[i] = i;
		}
	}
	for (y = 0; y < 9; y ++) {
		for (x = 0; x < 9; x ++) {
			b->i[y][x] = n[b->i[y][x]];
		}
	}
	make_region_index(b);
	if (check_region(b)) return FALSE;
	for (i = 0; i < 9 * 9 && *p; p ++) {
		if (*p == _T('0')) {
			i ++;
		} else if (*p >= _T('1') && *p <= _T('9')) {
			if (put_number(b, i % 9, i / 9, *p - _T('0'))) {
				break;
			}
			i ++;
		}
	}
	if (i != 81) clear_number(b);
	return TRUE;
}

#endif
