#include <stdio.h>
#include <stdarg.h>
#include <windows.h>
#include <commctrl.h>
#include <tchar.h>

void trace(LPTSTR lpFormat, ...)
{
#ifndef NDEBUG
	TCHAR lpMessage[256];
	va_list ap;
	
	va_start(ap, lpFormat);
	_vsntprintf(lpMessage, 256, lpFormat, ap);
	va_end(ap);
	OutputDebugString(lpMessage);
#endif
}

LPTSTR getclipboard(HWND hwnd)
{
	HGLOBAL hglobal;
	LPTSTR psz = NULL, str = NULL;
	int size;
	
	if (OpenClipboard(hwnd)) {
#if _UNICODE
		hglobal = GetClipboardData(CF_UNICODETEXT);
#else
		hglobal = GetClipboardData(CF_TEXT);
#endif
		if (hglobal != NULL) {
			size = GlobalSize(hglobal);
			psz = (LPTSTR) GlobalLock(hglobal);
			if (psz) {
				str = malloc(size + sizeof(TCHAR));
				if (str) {
					memcpy(str, psz, size);
					str[size] = _T('\0');
				}
				GlobalUnlock(hglobal);
			}
		}
		CloseClipboard();
	}
	return str;
}

BOOL setclipboard_handle(HWND hwnd, const LPVOID data, UINT uFormat)
{
	BOOL result = TRUE;
	
	if (OpenClipboard(hwnd)) {
		if (EmptyClipboard()) {
			result = (SetClipboardData(uFormat, data) == NULL);
		}
		CloseClipboard();
	}
	return result;
}

BOOL setclipboard(HWND hwnd, LPCTSTR str)
{
	HGLOBAL hglobal;
	LPTSTR psz;
	int size;
	
	size = lstrlen(str);
	if (size ++ == 0) return TRUE;
	hglobal = GlobalAlloc(GHND, sizeof(TCHAR) * (size));
	if (hglobal == NULL) return TRUE;
	psz = (LPTSTR) GlobalLock(hglobal);
	if (psz && lstrcpy(psz, str)) {
		GlobalUnlock(hglobal);
#if _UNICODE
		if (! setclipboard_handle(hwnd, hglobal, CF_UNICODETEXT)) {
#else
		if (! setclipboard_handle(hwnd, hglobal, CF_TEXT)) {
#endif
			return FALSE;
		}
	}
	GlobalFree(hglobal);
	return TRUE;
}

HDC create_window_enh_metafile(HWND hwnd, LPCTSTR filename,
                                RECT *rc, LPCTSTR description)
{
	int hsize, vsize, hres, vres, x, y;
	HDC hdc, hdcMeta = NULL;
	
	hdc = GetDC(hwnd);
	if (hdc) {
		hsize = GetDeviceCaps(hdc, HORZSIZE);
		vsize = GetDeviceCaps(hdc, VERTSIZE);
		hres = GetDeviceCaps(hdc, HORZRES);
		vres = GetDeviceCaps(hdc, VERTRES);
		x = (hsize * 100) / hres;
		y = (vsize * 100) / vres;
		rc->left *= x;
		rc->right *= x;
		rc->top *= y;
		rc->bottom *= y;
		hdcMeta = CreateEnhMetaFile(hdc, filename, rc, description);
		ReleaseDC(hwnd, hdc);
	}
	return hdcMeta;
}

LPTSTR path_add_slash(LPTSTR lpPath)
{
	DWORD length;
	LPTSTR p;
	
	length = lstrlen(lpPath);
	p = CharPrev(lpPath, lpPath + length);
	if (length > 1 && *p != _T('\\')) {
		lstrcat(lpPath, "\\");
	}
	return lpPath;
}

LPTSTR get_profile_path(LPCTSTR name)
{
	LPTSTR lpPath;
	DWORD size;
	
	size = GetCurrentDirectory(0, NULL) + lstrlen(name) + 2;
	lpPath = malloc(sizeof(TCHAR) * size);
	if (! lpPath) return NULL;
	if (! GetCurrentDirectory(size, lpPath)) {
		free(lpPath);
		return NULL;
	}
	path_add_slash(lpPath);
	lstrcat(lpPath, name);
	return lpPath;
}

LPCTSTR get_basename(LPCTSTR lpPath)
{
	DWORD length;
	LPTSTR p;
	
	length = lstrlen(lpPath);
	p = CharPrev(lpPath, lpPath + length);
	while (*p != _T('\\')) {
		if (p == lpPath) return lpPath;
		p = CharPrev(lpPath, p);
	}
	return CharNext(p);
}

void set_statusbar_text(HWND hwnd, LPCTSTR lpFormat, ...)
{
	TCHAR lpMessage[127];
	va_list ap;
	
	va_start(ap, lpFormat);
	_vsntprintf(lpMessage, 127, lpFormat, ap);
	va_end(ap);
	SendMessage(hwnd, SB_SETTEXT, 0, (LPARAM) lpMessage);
}

void set_statusbar_text_id(HINSTANCE hinst, HWND hwnd, UINT id, ...)
{
	TCHAR lpMessage[127], lpFormat[127];
	va_list ap;
	
	if (LoadString(hinst, id, lpFormat, 127)) {
		va_start(ap, id);
		_vsntprintf(lpMessage, 127, lpFormat, ap);
		va_end(ap);
		SendMessage(hwnd, SB_SETTEXT, 0, (LPARAM) lpMessage);
	}
}

void set_window_text(HWND hwnd, LPCTSTR lpFormat, ...)
{
	TCHAR lpMessage[127];
	va_list ap;
	
	va_start(ap, lpFormat);
	_vsntprintf(lpMessage, 127, lpFormat, ap);
	va_end(ap);
	SetWindowText(hwnd, lpMessage);
}

void set_window_text_id(HINSTANCE hinst, HWND hwnd, UINT id, ...)
{
	TCHAR lpMessage[127], lpFormat[127];
	va_list ap;
	
	if (LoadString(hinst, id, lpFormat, 127)) {
		va_start(ap, id);
		_vsntprintf(lpMessage, 127, lpFormat, ap);
		va_end(ap);
		SetWindowText(hwnd, lpMessage);
	}
}

BOOL set_dlg_item_text_id(HINSTANCE hinst, HWND hwnd,
                          int idCtrl, UINT id, ...)
{
	TCHAR lpMessage[127], lpFormat[127];
	va_list ap;
	
	if (LoadString(hinst, id, lpFormat, 127)) {
		va_start(ap, id);
		_vsntprintf(lpMessage, 127, lpFormat, ap);
		va_end(ap);
		return SetDlgItemText(hwnd, idCtrl, lpMessage);
	}
	return FALSE;
}

int message_box_id(HINSTANCE hinst, HWND hwnd, UINT id, LPCTSTR title, UINT style, ...)
{
	TCHAR lpMessage[256], lpFormat[256];
	va_list ap;
	
	if (LoadString(hinst, id, lpFormat, 256)) {
		va_start(ap, style);
		_vsntprintf(lpMessage, 256, lpFormat, ap);
		va_end(ap);
		return MessageBoxEx(hwnd, lpMessage, title, style, LANG_NEUTRAL);
	}
	return 0;
}

void get_current_objects(HDC hdc, int n, UINT uTypes[], HGDIOBJ hObjs[])
{
	int i;
	for (i = 0; i < n; i ++) {
		hObjs[i] = GetCurrentObject(hdc, uTypes[i]);
	}
}

void set_current_objects(HDC hdc, int n, HGDIOBJ hObjs[])
{
	int i;
	for (i = 0; i < n; i ++) {
		if (hObjs[i]) {
			SelectObject(hdc, hObjs[i]);
		}
	}
}

void center_on_parent_window(HWND hwnd)
{
	RECT rcParent, rc;
	HWND hwndParent;
	int x, y;
	
	hwndParent = GetParent(hwnd);
	if (! hwndParent ||
	    ! GetWindowRect(hwndParent, &rcParent) ||
	    ! GetWindowRect(hwnd, &rc)) {
		return;
	}
	x = rcParent.left +
	    (rcParent.right - rcParent.left - rc.right + rc.left) / 2;
	y = rcParent.top +
		(rcParent.bottom - rcParent.top - rc.bottom + rc.top) / 2;
	SetWindowPos(hwnd, 0, x, y, 0, 0, SWP_NOSIZE | SWP_NOZORDER);
}
